@@ -41,12 +41,15 @@
 #include "OpenImageIO/imageio.h"
 
 
-OIIO_NAMESPACE_ENTER
-{
+OIIO_NAMESPACE_BEGIN
+
+struct ROI;
 
 namespace pvt {
 // Forward declaration
 class ImageCacheImpl;
+class ImageCacheFile;
+class ImageCachePerThreadInfo;
 };
 
 
@@ -118,6 +121,47 @@ public:
     virtual bool getattribute (string_view name, char **val) = 0;
     virtual bool getattribute (string_view name, std::string &val) = 0;
 
+    /// Define an opaque data type that allows us to have a pointer
+    /// to certain per-thread information that the ImageCache maintains.
+    /// Any given one of these should NEVER be shared between running
+    /// threads.
+    typedef pvt::ImageCachePerThreadInfo Perthread;
+
+    /// Retrieve a Perthread, unique to the calling thread. This is a
+    /// thread-specific pointer that will always return the Perthread for a
+    /// thread, which will also be automatically destroyed when the thread
+    /// terminates.
+    ///
+    /// Applications that want to manage their own Perthread pointers (with
+    /// create_thread_info and destroy_thread_info) should still call this,
+    /// but passing in their managed pointer. If the passed-in thread_info
+    /// is not NULL, it won't create a new one or retrieve a TSP, but it
+    /// will do other necessary housekeeping on the Perthread information.
+    virtual Perthread * get_perthread_info (Perthread *thread_info = NULL) = 0;
+
+    /// Create a new Perthread. It is the caller's responsibility to
+    /// eventually destroy it using destroy_thread_info().
+    virtual Perthread * create_thread_info () = 0;
+
+    /// Destroy a Perthread that was allocated by create_thread_info().
+    virtual void destroy_thread_info (Perthread *thread_info) = 0;
+
+    /// Define an opaque data type that allows us to have a handle to an
+    /// image (already having its name resolved) but without exposing
+    /// any internals.
+    typedef pvt::ImageCacheFile ImageHandle;
+
+    /// Retrieve an opaque handle for fast image lookups.  The opaque
+    /// pointer thread_info is thread-specific information returned by
+    /// get_perthread_info().  Return NULL if something has gone
+    /// horribly wrong.
+    virtual ImageHandle * get_image_handle (ustring filename,
+                                            Perthread *thread_info=NULL) = 0;
+
+    /// Return true if the image handle (previously returned by
+    /// get_image_handle()) is a valid image that can be subsequently read.
+    virtual bool good (ImageHandle *file) = 0;
+
     /// Given possibly-relative 'filename', resolve it using the search
     /// path rules and return the full resolved filename.
     virtual std::string resolve_filename (const std::string &filename) const=0;
@@ -128,12 +172,9 @@ public:
     /// doesn't match the type requested. or some other failure.
     virtual bool get_image_info (ustring filename, int subimage, int miplevel,
                          ustring dataname, TypeDesc datatype, void *data) = 0;
-
-    /// Back-compatible version of get_image_info -- DEPRECATED
-    bool get_image_info (ustring filename, ustring dataname,
-                         TypeDesc datatype, void *data) {
-        return get_image_info (filename, 0, 0, dataname, datatype, data);
-    }
+    virtual bool get_image_info (ImageHandle *file, Perthread *thread_info,
+                         int subimage, int miplevel,
+                         ustring dataname, TypeDesc datatype, void *data) = 0;
 
     /// Get the ImageSpec associated with the named image (the first
     /// subimage & miplevel by default, or as set by 'subimage' and
@@ -144,6 +185,10 @@ public:
     virtual bool get_imagespec (ustring filename, ImageSpec &spec,
                                 int subimage=0, int miplevel=0,
                                 bool native=false) = 0;
+    virtual bool get_imagespec (ImageHandle *file, Perthread *thread_info,
+                                ImageSpec &spec,
+                                int subimage=0, int miplevel=0,
+                                bool native=false) = 0;
 
     /// Return a pointer to an ImageSpec associated with the named image
     /// (the first subimage & miplevel by default, or as set by
@@ -158,6 +203,10 @@ public:
     /// file, or invalidate_all(), or destroys the ImageCache.
     virtual const ImageSpec *imagespec (ustring filename, int subimage=0,
                                         int miplevel=0, bool native=false) = 0;
+    virtual const ImageSpec *imagespec (ImageHandle *file,
+                                        Perthread *thread_info,
+                                        int subimage=0, int miplevel=0,
+                                        bool native=false) = 0;
 
     /// Retrieve the rectangle of pixels spanning [xbegin..xend) X
     /// [ybegin..yend) X [zbegin..zend), with "exclusive end" a la STL,
@@ -177,6 +226,11 @@ public:
                              int xbegin, int xend, int ybegin, int yend,
                              int zbegin, int zend,
                              TypeDesc format, void *result) = 0;
+    virtual bool get_pixels (ImageHandle *file, Perthread *thread_info,
+                             int subimage, int miplevel,
+                             int xbegin, int xend, int ybegin, int yend,
+                             int zbegin, int zend,
+                             TypeDesc format, void *result) = 0;
 
     /// Retrieve the rectangle of pixels spanning [xbegin..xend) X
     /// [ybegin..yend) X [zbegin..zend), channels [chbegin..chend), 
@@ -191,7 +245,9 @@ public:
     /// enough to accommodate the requested rectangle (taking into
     /// consideration its dimensions, number of channels, and data
     /// format).  Requested pixels outside the valid pixel data region
-    /// will be filled in with 0 values.
+    /// will be filled in with 0 values. The optional cache_chbegin and
+    /// cache_chend hint as to which range of channels should be cached
+    /// (which by default will be all channels of the file).
     ///
     /// Return true if the file is found and could be opened by an
     /// available ImageIO plugin, otherwise return false.
@@ -200,54 +256,103 @@ public:
                     int ybegin, int yend, int zbegin, int zend,
                     int chbegin, int chend, TypeDesc format, void *result,
                     stride_t xstride=AutoStride, stride_t ystride=AutoStride,
-                    stride_t zstride=AutoStride) = 0;
+                    stride_t zstride=AutoStride,
+                    int cache_chbegin = 0, int cache_chend = -1) = 0;
+    virtual bool get_pixels (ImageHandle *file, Perthread *thread_info,
+                    int subimage, int miplevel, int xbegin, int xend,
+                    int ybegin, int yend, int zbegin, int zend,
+                    int chbegin, int chend, TypeDesc format, void *result,
+                    stride_t xstride=AutoStride, stride_t ystride=AutoStride,
+                    stride_t zstride=AutoStride,
+                    int cache_chbegin = 0, int cache_chend = -1) = 0;
 
     /// Define an opaque data type that allows us to have a pointer
     /// to a tile but without exposing any internals.
     class Tile;
 
-    /// Find a tile given by an image filename, subimage & miplevel, and
-    /// pixel coordinates.  An opaque pointer to the tile will be
-    /// returned, or NULL if no such file (or tile within the file)
-    /// exists or can be read.  The tile will not be purged from the
-    /// cache until after release_tile() is called on the tile pointer
-    /// the same number of times that get_tile() was called (refcnt).
-    /// This is thread-safe!
+    /// Find a tile given by an image filename, subimage & miplevel, channel
+    /// range, and pixel coordinates.  An opaque pointer to the tile will be
+    /// returned, or NULL if no such file (or tile within the file) exists
+    /// or can be read.  The tile will not be purged from the cache until
+    /// after release_tile() is called on the tile pointer the same number
+    /// of times that get_tile() was called (refcnt). This is thread-safe!
+    /// If chend < chbegin, it will retrieve a tile containing all channels
+    /// in the file.
     virtual Tile * get_tile (ustring filename, int subimage, int miplevel,
-                                int x, int y, int z) = 0;
+                             int x, int y, int z,
+                             int chbegin = 0, int chend = -1) = 0;
+    virtual Tile * get_tile (ImageHandle *file, Perthread *thread_info,
+                             int subimage, int miplevel,
+                             int x, int y, int z,
+                             int chbegin = 0, int chend = -1) = 0;
 
     /// After finishing with a tile, release_tile will allow it to
     /// once again be purged from the tile cache if required.
     virtual void release_tile (Tile *tile) const = 0;
 
+    /// Retrieve the data type of the pixels stored in the tile, which may
+    /// be different than the type of the pixels in the disk file.
+    virtual TypeDesc tile_format (const Tile *tile) const = 0;
+
+    /// Retrieve the ROI describing the pixels and channels stored in the
+    /// tile.
+    virtual ROI tile_roi (const Tile *tile) const = 0;
+
     /// For a tile retrived by get_tile(), return a pointer to the
     /// pixel data itself, and also store in 'format' the data type that
     /// the pixels are internally stored in (which may be different than
     /// the data type of the pixels in the disk file).
     virtual const void * tile_pixels (Tile *tile, TypeDesc &format) const = 0;
 
-    /// This creates a file entry in the cache that, instead of reading
-    /// from disk, uses a custom ImageInput to generate the image (note
-    /// that it will have no effect if there's already an image by the
-    /// same name in the cache).  The 'creator' is a factory that
-    /// creates the custom ImageInput and will be called like this:
-    /// ImageInput *in = creator(); Once created, the ImageCache owns
-    /// the ImageInput and is responsible for destroying it when done.
-    /// Custom ImageInputs allow "procedural" images, among other
-    /// things.  Also, this is the method you use to set up a
-    /// "writeable" ImageCache images (perhaps with a type of ImageInput
-    /// that's just a stub that does as little as possible).
-    virtual bool add_file (ustring filename, ImageInput::Creator creator) = 0;
+    /// The add_file() call causes a file to be opened or added to the
+    /// cache. There is no reason to use this method unless you are
+    /// supplying a custom creator, or configuration, or both.
+    /// 
+    /// If creator is not NULL, it points to an ImageInput::Creator that
+    /// will be used rather than the default ImageInput::create(), thus
+    /// instead of reading from disk, creates and uses a custom ImageInput
+    /// to generate the image. The 'creator' is a factory that creates the
+    /// custom ImageInput and will be called like this:
+    ///      ImageInput *in = creator();
+    /// Once created, the ImageCache owns the ImageInput and is responsible
+    /// for destroying it when done. Custom ImageInputs allow "procedural"
+    /// images, among other things.  Also, this is the method you use to set
+    /// up a "writeable" ImageCache images (perhaps with a type of
+    /// ImageInput that's just a stub that does as little as possible).
+    /// 
+    /// If config is not NULL, it points to an ImageSpec with configuration
+    /// options/hints that will be passed to the underlying
+    /// ImageInput::open() call. Thus, this can be used to ensure that the
+    /// ImageCache opens a call with special configuration options.
+    /// 
+    /// This call (including any custom creator or configuration hints) will
+    /// have no effect if there's already an image by the same name in the
+    /// cache. Custom creators or configurations only "work" the FIRST time
+    /// a particular filename is referenced in the lifetime of the
+    /// ImageCache.
+    virtual bool add_file (ustring filename, ImageInput::Creator creator=NULL,
+                           const ImageSpec *config=NULL) = 0;
 
     /// Preemptively add a tile corresponding to the named image, at the
-    /// given subimage and MIP level.  The tile added is the one whose
-    /// corner is (x,y,z), and buffer points to the pixels (in the given
-    /// format, with supplied strides) which will be copied and inserted
-    /// into the cache and made available for future lookups.
+    /// given subimage, MIP level, and channel range.  The tile added is the
+    /// one whose corner is (x,y,z), and buffer points to the pixels (in the
+    /// given format, with supplied strides) which will be copied and
+    /// inserted into the cache and made available for future lookups.
+    /// If chend < chbegin, it will add a tile containing the full set of
+    /// channels for the image.
+    virtual bool add_tile (ustring filename, int subimage, int miplevel,
+                     int x, int y, int z, int chbegin, int chend,
+                     TypeDesc format, const void *buffer,
+                     stride_t xstride=AutoStride, stride_t ystride=AutoStride,
+                     stride_t zstride=AutoStride) = 0;
+    // DEPRECATED(1.6) -- add a tile with all channels.
     virtual bool add_tile (ustring filename, int subimage, int miplevel,
                      int x, int y, int z, TypeDesc format, const void *buffer,
                      stride_t xstride=AutoStride, stride_t ystride=AutoStride,
-                     stride_t zstride=AutoStride) = 0;
+                     stride_t zstride=AutoStride) {
+        return add_tile (filename, subimage, miplevel, x, y, z, 0, -1,
+                         format, buffer, xstride, ystride, zstride);
+    }
 
     /// If any of the API routines returned false indicating an error,
     /// this routine will return the error string (and clear any error
@@ -296,7 +401,6 @@ private:
 };
 
 
-}
-OIIO_NAMESPACE_EXIT
+OIIO_NAMESPACE_END
 
 #endif // OPENIMAGEIO_IMAGECACHE_H
